#ifndef ROOTS_HH
#define ROOTS_HH
#include "exceptions.hh"
#include "lohi.hh"
#include <gmpxx.h>

namespace Interval {

    // An Indicant f, when invoked as f(x), returns
    //
    //   + a negative value if x lies too close to f.min() but
    //   + a positive value if x lies too far from f.min().
    //
    // Since x is not a mathematical real number but a fixed-point
    // approximation thereto, there may not exist an x for which f(x)
    // returns exactly zero.  However, if there does, then that x is
    // just right.
    //
    // A call to f.flank() returns, as the name suggests, the flank of
    // f.min() on which a root is to be sought -- -1 for the left flank
    // or +1 for the right.  If the call returns 0, then f.min() itself
    // is the root.
    //
    class Indicant {
    private:
        mp_bitcnt_t precision_;
    protected:
        long order_;
        bool prefer_negative_branch_;
    public:
        explicit Indicant(mp_bitcnt_t, long, bool = false);
        virtual ~Indicant() {}
        inline mp_bitcnt_t precision() const { return precision_; }
        inline long order() const { return order_; }
        inline bool prefer_negative_branch() const
          { return prefer_negative_branch_; }
        virtual mpz_class operator()(const mpz_class &) const = 0;
        virtual Lohi min() const = 0;
        virtual signed char flank() const = 0;
    };

    // Indicate against (p + s*sqrt(n)) / q.
    class Indicant_re_general_quadratic: public Indicant {
    private:
        Lohi min_;
        mpz_class a2;
        mpz_class a1;
        mpz_class a0;
        signed char flank_;
    public:
        Indicant_re_general_quadratic(
            mp_bitcnt_t precision, long n, long s, long p, long q,
            bool prefer_negative_branch = false
        );
        mpz_class operator()(const mpz_class &) const;
        Lohi min() const;
        signed char flank() const;
    };

    // Indicate against (p/q)^(n/m), in which, after reduction, m > 0.
    // If everything is positive, then f(x) = x^m * q^n - p^n.
    // If something is negative, then p and q may be swapped
    // and/or f(x) may be negated as necessary to cause f(x)
    // to indicate as it ought.
    class Indicant_re_power: public Indicant {
    private:
        mpz_class am;
        mpz_class a0;
        long m;
        signed char flank_;
    public:
        Indicant_re_power(
            mp_bitcnt_t precision, long n, long m, long p, long q,
            bool prefer_negative_branch = false
        );
        mpz_class operator()(const mpz_class &) const;
        Lohi min() const;
        signed char flank() const;
    };

    // Indicate against irrat^(n/m); f(x) = x^m - irrat^n.
    // (The irrat is not actually an irrational number, but a
    // precision-bit approximation to an irrational number or,
    // indeed, to any real number.)
    class Indicant_re_irrat_to_power: public Indicant {
        mpz_class coeff;
        mp_bitcnt_t log2c;
        long m;
        signed char flank_;
        bool was_n_negative;
        bool do_negate_f;
    public:
        Indicant_re_irrat_to_power(
            mp_bitcnt_t precision, long n, long m, mpz_class irrat,
            bool prefer_negative_branch = false
        );
        mpz_class operator()(const mpz_class &) const;
        Lohi min() const;
        signed char flank() const;
    };

    Lohi root(const Indicant &);
    Lohi root(
        mp_bitcnt_t precision, long n, long m, const Lohi &,
        bool prefer_negative_branch = false
    );

}

#endif

