#include "stringific.hh"
#include <iterator>
#include <algorithm>
#include <list>
#include <iomanip>
#include <sstream>

std::string Interval::stringify
  (mp_bitcnt_t precision, mpz_class n, const Stringify_style style)
{
    std::string result;
    constexpr size_t leftward_width {5 * no_of_leftward_groups};
    constexpr size_t no_of_rightward_digits_per_line
      {no_of_rightward_groups << 2};
    const std::string continuing_pad(leftward_width + 1, ' ');
    bool is_negative {false};
    if (n < 0) n = -n, is_negative = true;
    if (style == Stringify_style::integer) precision = 0;
    else {
        const mp_bitcnt_t bitpad_size
          {precision ? 3 - ((precision + 3) & 3) : 4};
        n <<= bitpad_size, precision += bitpad_size;
    }
    std::list<char> leftward_digits, rightward_digits;
    {
        // The writer would have preferred to compose the string
        // without resort to stream mechanics, but it is in streams
        // that the standard library handles unusual character sets
        // like EBCDIC.
        std::ostringstream leftward_stream, rightward_stream;
        leftward_stream << std::hex, rightward_stream << std::hex;
        while (precision) {
            const unsigned long fourbit {n.get_ui() & 0xf};
            n >>= 4, precision -= 4;
            rightward_stream << fourbit;
        }
        do {
            const unsigned long fourbit {n.get_ui() & 0xf};
            n >>= 4;
            leftward_stream << fourbit;
        } while (n);
        //
        // For some reason (probably for a good reason unknown to the
        // writer), before one iterates over a stringstream's string,
        // it seems to be necessary to copy or at least to move the
        // string out of the stream.  Apparently, one can iterate over
        // a copy -- even if move-constructed -- without trouble, but
        // not over the original.
        //
        // Whether moving the stream's string representation
        // effectively differs from copying it is unknown to
        // the writer.  The representation's inability to be iterated
        // over suggests that the representation lacks a coherent,
        // movement-capable string object.
        //
        // One could instead have used std::snprintf(), skipping the
        // stringstream altogether.  Nevertheless, balancing
        // contrary considerations of expected speed, expected power
        // usage, expected memory usage, maintainability and
        // aesthetics, the writer prefers the stringstream solution.
        //
        const std::string
          leftward_unstream  {std::move(leftward_stream .str())},
          rightward_unstream {std::move(rightward_stream.str())};
        // The streams have the digits reversed, so a
        // std::front_inserter is next used to dereverse them.
        std::copy(
            leftward_unstream.begin(),
            leftward_unstream.end(),
            std::front_inserter(leftward_digits)
        );
        std::copy(
            rightward_unstream.begin(),
            rightward_unstream.end(),
            std::front_inserter(rightward_digits)
        );
    }
    {
        std::string leftward_string {is_negative ? "-" : ""};
        {
            bool has_written {false};
            while (!leftward_digits.empty()) {
                if (!(leftward_digits.size() & 3) && has_written)
                  leftward_string += ' ';
                leftward_string += std::move(leftward_digits.front()),
                  leftward_digits.pop_front();
                has_written = true;
            }
        }
        result = std::string(
            (
                leftward_string.length() <= leftward_width
                ? leftward_width - leftward_string.length() : 0
            ), ' '
        );
        result += leftward_string;
        if (style != Stringify_style::integer) result += '.';
    }
    {
        size_t count {0};
        while (!rightward_digits.empty()) {
            if (count) {
                if (!(count % no_of_rightward_digits_per_line)) {
                    result += '\n';
                    result += continuing_pad;
                }
                else if (!(count & 3)) result += ' ';
            }
            result += std::move(rightward_digits.front()),
              rightward_digits.pop_front();
            ++count;
        }
    }
    result += '\n';
    return result;
}

